﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace tetris.v4
{
  public class Controller
  {
    public enum Directions
    {
      Left = 1,
      Right = 2,
      Down = 3,
      Rotate = 4,
      CheckStartPiece = 5
    }

    public bool testCanMove(Directions dir, Memory where)
    {
      switch (dir)
      {
        case Directions.Left:
          for (int j = 0; j < Memory.PEACE_SIZE_X; j++)
          {
            for (int i = 0; i < Memory.PEACE_SIZE_Y; i++)
            {
              if (where.currPiece[j, i])
              {
                int pointX = j + where.currX;
                if (pointX < 0 || pointX >= Memory.SCREEN_COLUMNS)
                {
                  return false;
                }
                int pointY = i + where.currY;
                if (pointY < 0 || pointY >= Memory.SCREEN_LINES)
                {
                  return false;
                }
                if (pointX > 0)
                {
                  if (where.screen[pointX - 1, pointY] != 0)
                  {
                    return false;
                  }
                }
                else
                {
                  return false;
                }
              }
            }
          }
          break;
        case Directions.Right:
          for (int j = 0; j < Memory.PEACE_SIZE_X; j++)
          {
            for (int i = 0; i < Memory.PEACE_SIZE_Y; i++)
            {
              if (where.currPiece[j, i])
              {
                int pointX = j + where.currX;
                if (pointX < 0 || pointX >= Memory.SCREEN_COLUMNS)
                {
                  return false;
                }
                int pointY = i + where.currY;
                if (pointY < 0 || pointY >= Memory.SCREEN_LINES)
                {
                  return false;
                }
                if (pointX < Memory.SCREEN_COLUMNS - 1)
                {
                  if (where.screen[pointX + 1, pointY] != 0)
                  {
                    return false;
                  }
                }
                else
                {
                  return false;
                }
              }
            }
          }
          break;
        case Directions.Down:
          for (int j = 0; j < Memory.PEACE_SIZE_X; j++)
          {
            for (int i = 0; i < Memory.PEACE_SIZE_Y; i++)
            {
              if (where.currPiece[j, i])
              {
                int pointX = j + where.currX;
                if (pointX < 0 || pointX >= Memory.SCREEN_COLUMNS)
                {
                  return false;
                }
                int pointY = i + where.currY;
                if (pointY < 0 || pointY >= Memory.SCREEN_LINES)
                {
                  return false;
                }
                if (pointY < Memory.SCREEN_LINES - 1)
                {
                  if (where.screen[pointX, pointY + 1] != 0)
                  {
                    return false;
                  }
                }
                else
                {
                  return false;
                }
              }
            }
          }
          break;
        case Directions.Rotate:
          for (int j = 0; j < Memory.PEACE_SIZE_X; j++)
          {
            for (int i = 0; i < Memory.PEACE_SIZE_Y; i++)
            {
              if (where.currPieceAfterRotate[j, i])
              {
                int pointX = j + where.currX;
                if (pointX < 0 || pointX >= Memory.SCREEN_COLUMNS)
                {
                  return false;
                }
                int pointY = i + where.currY;
                if (pointY < 0 || pointY >= Memory.SCREEN_LINES)
                {
                  return false;
                }
                if (where.screen[pointX, pointY] != 0)
                {
                  return false;
                }
              }
            }
          }
          break;
        case Directions.CheckStartPiece:
          for (int j = 0; j < Memory.PEACE_SIZE_X; j++)
          {
            for (int i = 0; i < Memory.PEACE_SIZE_Y; i++)
            {
              if (where.currPiece[j, i])
              {
                int pointX = j + where.currX;
                if (pointX < 0 || pointX >= Memory.SCREEN_COLUMNS)
                {
                  return false;
                }
                int pointY = i + where.currY;
                if (pointY < 0 || pointY >= Memory.SCREEN_LINES)
                {
                  return false;
                }
                if (where.screen[pointX, pointY] != 0)
                {
                  return false;
                }
              }
            }
          }
          break;
      }

      return true;
    }

    public void move(Directions dir, Memory where)
    {
      switch (dir)
      {
        case Directions.Left:
          where.currX--;
          break;
        case Directions.Right:
          where.currX++;
          break;
        case Directions.Down:
          where.currY++;
          break;
      }
    }

    public void pieceEnd(Memory where)
    {
      for (int j = 0; j < Memory.PEACE_SIZE_X; j++)
      {
        for (int i = 0; i < Memory.PEACE_SIZE_Y; i++)
        {
          if (where.currPiece[j, i])
          {
            int pointX = j + where.currX;
            int pointY = i + where.currY;
            if (pointX >= 0 && pointX < Memory.SCREEN_COLUMNS && pointY >= 0 && pointY < Memory.SCREEN_LINES)
            {
              where.screen[pointX, pointY] = (int)where.currPieceType;
            }
            where.currPiece[j, i] = false;
          }
        }
      }
    }

    public bool checkLine(Memory where, int y)
    {
      for (int j = 0; j < Memory.SCREEN_COLUMNS; j++)
      {
        if (where.screen[j, y] == 0)
        {
          return false;
        }
      }
      return true;
    }

    public void reomveLine(Memory where, int y)
    {
      if (y > 0)
      {
        for (int i = y - 1; i >= 0; i--)
        {
          for (int j = 0; j < Memory.SCREEN_COLUMNS; j++)
          {
            where.screen[j, i + 1] = where.screen[j, i];
          }
        }
      }
      for (int j = 0; j < Memory.SCREEN_COLUMNS; j++)
      {
        where.screen[j, 0] = 0;
      }
    }

    public void flashLine(Memory where, int y, int color)
    {
      for (int j = 0; j < Memory.SCREEN_COLUMNS; j++)
      {
        where.screen[j, y] = color;
      }
    }

    public void applyRotate(Memory where)
    {

      for (int j = 0; j < Memory.PEACE_SIZE_X; j++)
      {
        for (int i = 0; i < Memory.PEACE_SIZE_Y; i++)
        {
          where.currPiece[j, i] = where.currPieceAfterRotate[j, i];
        }
      }
      int currRotation = (int)where.currPieceRotation;
      int newRotation = currRotation < 4 ? currRotation + 1 : 1;
      where.currPieceRotation = (Memory.pieceRotation)newRotation;
    }

    public void rotateIfPossbile(Memory where)
    {

      //for (int j = 0; j < Memory.PEACE_SIZE_X; j++)
      //{
      //  for (int i = 0; i < Memory.PEACE_SIZE_Y; i++)
      //  {
      //    double pi = 3.14159265;
      //    double radiantsRotation = 90 * pi / 180;
      //    int newX = (int)(j * Math.Cos(radiantsRotation) + i * Math.Sin(radiantsRotation));
      //    int newY = (int)(i * Math.Cos(radiantsRotation) - j * Math.Sin(radiantsRotation));
      //    if (newX >= 0 && newX < Memory.PEACE_SIZE_X && newY >= 0 && newY < Memory.PEACE_SIZE_Y)
      //    {
      //      where.currPieceAfterRotate[newX, newY] = where.currPiece[j, i];
      //    }
      //  }
      //}
      int currRotation = (int)where.currPieceRotation;
      int newRotation = currRotation < 4 ? currRotation + 1 : 1;
      where.currPieceAfterRotate = where.getPiece(where.currPieceType, (Memory.pieceRotation)newRotation);
      
      if (this.testCanMove(Directions.Rotate, where))
      {
        this.applyRotate(where);
      }
    }


    public void resetMemoryArrays(Memory where)
    {
      for (int i = 0; i < Memory.SCREEN_LINES; i++)
      {
        for (int J = 0; J < Memory.SCREEN_COLUMNS; J++)
        {
          where.screen[J, i] = 0;
        }
      }
      for (int j = 0; j < Memory.PEACE_SIZE_X; j++)
      {
        for (int i = 0; i < Memory.PEACE_SIZE_Y; i++)
        {
          where.currPiece[j, i] = false;
          where.next[j, i] = false;
          where.currPieceAfterRotate[j, i] = false;
        }
      }
    }

  }
}


