﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.IO.Compression;
using System.Linq;
using System.Net;
using System.Reflection;
using System.Text;
using System.Windows.Forms;

namespace tetris.v4
{
  public partial class gamePanel : Form
  {
    private Memory tetrisGameMap = new Memory();
    private Controller controller = new Controller();
    private int currPieceSize = 0;
    string currkey = null;
    bool isRelasedPiece = false;
    bool isArrived = false;
    bool checkingLines = false;
    bool checkLinesEnded = true;
    int totalLineDone = 0;
    int lineToFlash = 0;
    bool isFlashing = false;
    int currFlashStatus = 0;
    DateTime lastMove = DateTime.Now;
    DateTime lastFlash = DateTime.Now;
    public int currTick = 0;
    public List<Recorder> recordGame = new List<Recorder>();
    bool isDemoMode = false;
    bool isGameOver = false;
    bool record = false;
    public gamePanel()
    {
      InitializeComponent();
      controller.resetMemoryArrays(tetrisGameMap);
      timer1.Interval = 10;

    }

    private void gamePanel_Load(object sender, EventArgs e)
    {
      gamePanel_Resize(null, null);
    }

    private void nextPiece_Paint(object sender, PaintEventArgs e)
    {
      var graphics = e.Graphics;

      Brush txtB = new SolidBrush(Color.Black);
      string txtNext = "NEXT:";
      Font nextFont = new Font("Verdana", 12f);
      graphics.DrawString(txtNext, nextFont, txtB, 45, 10);

      Brush b = new SolidBrush(tetrisGameMap.pieceType2Color(tetrisGameMap.nextType));
      Pen p = new Pen(Color.Black);
      p.Width = 2;


      for (int i = 0; i < Memory.PEACE_SIZE_Y; i++)
      {
        for (int j = 0; j < Memory.PEACE_SIZE_X; j++)
        {
          if (tetrisGameMap.next[j, i])
          {
            int x1 = j * 20 + 50;
            int y1 = i * 20 + 50;
            int x2 = x1 + 20;
            int y2 = y1 + 20;
            graphics.FillRectangle(b, x1, y1, 20, 20);
            graphics.DrawRectangle(p, x1, y1, 20, 20);
          }
        }
      }
    }

    private void score_Paint(object sender, PaintEventArgs e)
    {
      var graphics = e.Graphics;
      Brush txtB = new SolidBrush(Color.Black);
      string txtLines = string.Format("Lines: {0}", tetrisGameMap.currLines); 
      string txtScore = string.Format("Score: {0}", tetrisGameMap.currScore); 
      string txtLevel = string.Format("Level: {0}", Memory.LEVEL_OFFSET - tetrisGameMap.currLevel);
      Font infoFont = new Font("Courier new", 15f);
      graphics.DrawString(txtLines, infoFont, txtB, 15, 20);
      graphics.DrawString(txtScore, infoFont, txtB, 15, 50);
      graphics.DrawString(txtLevel, infoFont, txtB, 15, 80);
      Font buttonFont = new Font("Verdana", 8f);
      string text = "Left move left\nRight move right\nUp increase level\nDown release piece\nEnter rotate";
      //SizeF textSize = graphics.MeasureString(text, buttonFont);
      graphics.DrawString(text, buttonFont, txtB, 15, 110);

    }

    private void piecesPanel_Paint(object sender, PaintEventArgs e)
    {
      var graphics = e.Graphics;
      Pen p = new Pen(Color.Black);
      p.Width = 2;


      for (int i = 0; i < Memory.SCREEN_LINES; i++)
      {
        for (int j = 0; j < Memory.SCREEN_COLUMNS; j++)
        {
          int color = tetrisGameMap.screen[j, i];
          if (color != 0)
          {
            Brush b = new SolidBrush(tetrisGameMap.pieceType2Color((Memory.pieceType) color));
            int x1 = j * currPieceSize;
            int y1 = i * currPieceSize;
            graphics.FillRectangle(b, x1, y1, currPieceSize, currPieceSize);
            graphics.DrawRectangle(p, x1, y1, currPieceSize, currPieceSize);
          }
        }
      }

      Brush brushCurrPiece = new SolidBrush(tetrisGameMap.pieceType2Color(tetrisGameMap.currPieceType));
      for (int i = 0; i < Memory.PEACE_SIZE_Y; i++)
      {
        for (int j = 0; j < Memory.PEACE_SIZE_X; j++)
        {
          if (tetrisGameMap.currPiece[j, i])
          {
            int x1 = (j + tetrisGameMap.currX) * currPieceSize;
            int y1 = (i + tetrisGameMap.currY) * currPieceSize;
            graphics.FillRectangle(brushCurrPiece, x1, y1, currPieceSize, currPieceSize);
            graphics.DrawRectangle(p, x1, y1, currPieceSize, currPieceSize);
          }
        }
      }

      if (isDemoMode || isGameOver)
      {
        Brush txtB = null;
        string txt = null;

        if (isDemoMode )
        {
          txtB = new SolidBrush(Color.Orange);
          txt = "DEMO";
        }
        else if (isGameOver)
        {
          txtB = new SolidBrush(Color.Red);
          txt = "GAME OVER";
        }
        
        Font txtFont = new Font("Verdana", 15f);
        SizeF textSize = graphics.MeasureString(txt, txtFont);
        int x1 = (int)(piecesPanel.Width - textSize.Width) / 2;
        int y1 = (int)(piecesPanel.Height - textSize.Height) / 2 - 100;
        graphics.DrawString(txt, txtFont, txtB, x1, y1);
      }

    }


    private void panelLogo_Paint(object sender, PaintEventArgs e)
    {
      var graphics = e.Graphics;
      Brush b = new SolidBrush(Color.Blue);
      graphics.FillRectangle(b, new Rectangle(0, 0, panelLogo.Width, panelLogo.Height));
      Brush txtB = new SolidBrush(Color.Yellow);
      string logo = "Tetris by Antoine 2019\n"+
                    "    Dedicated 2 Valentina";
      Font tetrisFont = new Font("Verdana", 8f);
      SizeF textSize = graphics.MeasureString(logo, tetrisFont);
      graphics.DrawString(logo, tetrisFont, txtB, (panelLogo.Width - textSize.Width) / 2, (panelLogo.Height - textSize.Height) / 2);
    }

    private void gamePanel_Resize(object sender, EventArgs e)
    {
      bool fit = false;
      int reduceFactor = 50;
      while (!fit)
      {
        int sizePanelbyHeight = panelGameArea.Height - reduceFactor ;
        int widthByHeight = ((int)(sizePanelbyHeight / Memory.SCREEN_LINES)) * Memory.SCREEN_COLUMNS;
        int sizePanelbyWidth = (panelGameArea.Width - reduceFactor );
        int heightByWidth = ((int)(sizePanelbyWidth / Memory.SCREEN_COLUMNS)) * Memory.SCREEN_LINES;

        int widthToUse = 0;
        int heightToUse = 0;
        if  (heightByWidth > sizePanelbyHeight )
        {
          heightToUse = heightByWidth;
          widthToUse = sizePanelbyWidth;
        }
        else
        {
          heightToUse = sizePanelbyHeight;
          widthToUse = widthByHeight;
        }

        if (heightToUse < panelGameArea.Height && widthToUse < panelGameArea.Width)
        {
          piecesPanel.Width = widthToUse;
          piecesPanel.Height = heightToUse;
          fit = true;
        }
        else
        {
          reduceFactor += 50;
        }
      }
      piecesPanel.Left = (panelGameArea.Width - piecesPanel.Width) / 2;
      piecesPanel.Top = panelGameArea.Top + (panelGameArea.Height - piecesPanel.Height) / 2;

      currPieceSize = piecesPanel.Height / Memory.SCREEN_LINES;

      piecesPanel.Invalidate();
      score.Invalidate();
      nextPiece.Invalidate();
      panelLogo.Invalidate();
    }

    private void gamePanel_KeyDown(object sender, KeyEventArgs e)
    {
      switch (e.KeyCode)
      {
        case Keys.Up:
          currkey = "up";
          break;
        case Keys.Down:
          currkey = "down";
          break;
        case Keys.Left:
          currkey = "left";
          break;
        case Keys.Right:
          currkey = "right";
          break;
        case Keys.Enter:
          currkey = "enter";
          break;
        default:
          currkey = null;
          break;
      }
    }

    private void cmdStart_Click(object sender, EventArgs e)
    {
      tetrisGameMap.currLevel = Memory.LEVEL_FIRST;
      tetrisGameMap.currLines = 0;
      tetrisGameMap.currScore = 0;
      controller.resetMemoryArrays(tetrisGameMap);
      tetrisGameMap.currX = (Memory.SCREEN_COLUMNS - Memory.PEACE_SIZE_X) / 2;
      tetrisGameMap.currY = 0;
      tetrisGameMap.currPieceType = tetrisGameMap.getRandomPiece();
      tetrisGameMap.currPieceRotation = Memory.pieceRotation.NO;
      System.Threading.Thread.Sleep(1000);
      tetrisGameMap.nextType = tetrisGameMap.getRandomPiece();
      tetrisGameMap.currPiece = (bool[,])tetrisGameMap.getPiece(tetrisGameMap.currPieceType, tetrisGameMap.currPieceRotation).Clone();
      tetrisGameMap.next = (bool[,])tetrisGameMap.getPiece(tetrisGameMap.nextType, Memory.pieceRotation.NO).Clone();

      piecesPanel.Invalidate();
      score.Invalidate();
      nextPiece.Invalidate();
      panelLogo.Invalidate();
      
      piecesPanel.Focus();
      isRelasedPiece = false;
      isArrived =false;
      checkingLines = false;
      checkLinesEnded = true;
      timer1.Enabled = true;
      lastMove = DateTime.Now;
      currTick = 0;
      recordGame = new List<Recorder>();
      isDemoMode = false;
      isGameOver = false;
    }

    private void timer1_Tick(object sender, EventArgs e)
    {
      timer1.Enabled = false;
      currTick++;


      if (isDemoMode)
      {
        Recorder tick = (from T in recordGame where T.currTick == currTick select T).FirstOrDefault();
        if (tick == null)
        {
          isDemoMode = false;
          return;
        }

        if (tick.screen != null)
        {
          tetrisGameMap.screen = (int[,])tick.screen.Clone();
        }

        tetrisGameMap.currPiece = (bool[,])tick.currPiece.Clone();
        tetrisGameMap.currPieceType = tick.currPieceType;
        tetrisGameMap.next = (bool[,])tick.next.Clone();
        tetrisGameMap.nextType  = tick.nextType;
        tetrisGameMap.currX = tick.currX;
        tetrisGameMap.currY = tick.currY;
        tetrisGameMap.currLevel = tick.currLevel;
        tetrisGameMap.currLines = tick.currLines;
        tetrisGameMap.currScore = tick.currScore;
        if (tick.currOp == Recorder.Operation.canghedNext || currTick == 1)
        {
          score.Invalidate();
          nextPiece.Invalidate();
        }
        piecesPanel.Invalidate();
        timer1.Enabled = true;
        return;
      }

      if (record)
      {
        if (recordGame.Count() == 0)
        {
          Recorder tick1 = new Recorder();
          tick1.currOp = Recorder.Operation.noOp;
          tick1.screen = (int[,])tetrisGameMap.screen.Clone();
          tick1.currPiece = (bool[,])tetrisGameMap.currPiece.Clone();
          tick1.next = (bool[,])tetrisGameMap.next.Clone();
          tick1.currPieceType = tetrisGameMap.currPieceType;
          tick1.nextType = tetrisGameMap.nextType;
          tick1.currTick = currTick;
          tick1.currX = tetrisGameMap.currX;
          tick1.currY = tetrisGameMap.currY;
          tick1.currLevel = tetrisGameMap.currLevel;
          tick1.currLines = tetrisGameMap.currLines;
          tick1.currScore = tetrisGameMap.currScore;
          recordGame.Add(tick1);
        }
        else
        {
          var maxPrevTick = (from T in recordGame where T.screen != null select T.currTick).Max();
          Recorder tickPrev = (from T in recordGame where T.currTick == maxPrevTick select T).FirstOrDefault();
          Recorder newTick = new Recorder();
          if (newTick.screenIsCanghed(tickPrev.screen, tetrisGameMap))
          {
            newTick.screen = (int[,])tetrisGameMap.screen.Clone();
          }
          newTick.currPiece = (bool[,])tetrisGameMap.currPiece.Clone();
          newTick.next = (bool[,])tetrisGameMap.next.Clone();
          newTick.currPieceType = tetrisGameMap.currPieceType;
          newTick.nextType = tetrisGameMap.nextType;

          tickPrev = (from T in recordGame where T.currTick == currTick - 1 select T).FirstOrDefault();
          if (newTick.nextIsCanghed(tickPrev.next, tetrisGameMap))
          {
            newTick.currOp = Recorder.Operation.canghedNext;
          }
          else
          {
            newTick.currOp = Recorder.Operation.noOp;
          }
          newTick.currTick = currTick;
          newTick.currX = tetrisGameMap.currX;
          newTick.currY = tetrisGameMap.currY;
          newTick.currLevel = tetrisGameMap.currLevel;
          newTick.currLines = tetrisGameMap.currLines;
          newTick.currScore = tetrisGameMap.currScore;
          recordGame.Add(newTick);
        }
      }

      if (!string.IsNullOrEmpty(currkey))
      {
        switch (currkey)
        {
          case "left":
            if (controller.testCanMove(Controller.Directions.Left, tetrisGameMap))
            {
              controller.move(Controller.Directions.Left, tetrisGameMap);
              piecesPanel.Invalidate();
            }
            break;
          case "right":
            if (controller.testCanMove(Controller.Directions.Right, tetrisGameMap))
            {
              controller.move(Controller.Directions.Right, tetrisGameMap);
              piecesPanel.Invalidate();
            }
            break;
          case "up":
            if (tetrisGameMap.currLevel > 0)
            {
              tetrisGameMap.currLevel--;
            }
            score.Invalidate();
            break;
          case "down":
            isRelasedPiece = true;
            break;
          case "enter":
            controller.rotateIfPossbile(tetrisGameMap);
            piecesPanel.Invalidate();
            break;
        }
        currkey = null;
      }
      if (!checkingLines)
      {
        TimeSpan timeSpent = DateTime.Now - lastMove;
        if (isRelasedPiece || (int)timeSpent.TotalMilliseconds >= tetrisGameMap.currLevel * Memory.LEVEL_MILLISECOND_MULTIPLY)
        {
          if (controller.testCanMove(Controller.Directions.Down, tetrisGameMap))
          {
            lastMove = DateTime.Now;
            controller.move(Controller.Directions.Down, tetrisGameMap);
            piecesPanel.Invalidate();
            isArrived = false;
          }
          else
          {
            isRelasedPiece = false;
            isArrived = true;
          }
        }
        TimeSpan timeSpent2 = DateTime.Now - lastMove;
        if (isArrived && (int)timeSpent2.TotalMilliseconds >= tetrisGameMap.currLevel * Memory.LEVEL_MILLISECOND_MULTIPLY)
        {
          controller.pieceEnd(tetrisGameMap);
          checkingLines = true;
          checkLinesEnded = false;
          totalLineDone = 0;
          lineToFlash = 0;
          isFlashing = false;
        }
      }
      else
      {
        if (!checkLinesEnded)
        {
          if (isFlashing)
          {
            TimeSpan timeSpent = DateTime.Now - lastFlash;
            if ((int)timeSpent.TotalMilliseconds >= 80)
            {
              switch (currFlashStatus)
              {
                case 0:
                  controller.flashLine(tetrisGameMap, lineToFlash, (int)Memory.pieceType.LINEA_LAMPEGGIANTE_COL1);
                  piecesPanel.Invalidate();
                  currFlashStatus++;
                  lastFlash = DateTime.Now;
                  break;
                case 1:
                  controller.flashLine(tetrisGameMap, lineToFlash, (int)Memory.pieceType.LINEA_LAMPEGGIANTE_COL2);
                  piecesPanel.Invalidate();
                  currFlashStatus++;
                  lastFlash = DateTime.Now;
                  break;
                case 2:
                  controller.flashLine(tetrisGameMap, lineToFlash, (int)Memory.pieceType.LINEA_LAMPEGGIANTE_COL1);
                  piecesPanel.Invalidate();
                  currFlashStatus++;
                  lastFlash = DateTime.Now;
                  break;
                case 3:
                  controller.flashLine(tetrisGameMap, lineToFlash, (int)Memory.pieceType.LINEA_LAMPEGGIANTE_COL2);
                  piecesPanel.Invalidate();
                  currFlashStatus++;
                  lastFlash = DateTime.Now;
                  break;
                case 4:
                  controller.reomveLine(tetrisGameMap, lineToFlash);
                  piecesPanel.Invalidate();
                  isFlashing = false;
                  break;
              }
            }
          }
          else
          {
            for (int i = Memory.SCREEN_LINES - 1; i >= 0; i--)
            {
              if (controller.checkLine(tetrisGameMap, i))
              {
                isFlashing = true;
                lineToFlash = i;
                currFlashStatus = 0;
                lastFlash = DateTime.Now;
                tetrisGameMap.currLines++;
                if ((tetrisGameMap.currLines == 15
                  || tetrisGameMap.currLines == 28
                  || tetrisGameMap.currLines == 39
                  || tetrisGameMap.currLines == 48
                  || tetrisGameMap.currLines == 56
                  || tetrisGameMap.currLines == 63
                  || tetrisGameMap.currLines == 67
                  || tetrisGameMap.currLines == 78
                  || tetrisGameMap.currLines == 92) && tetrisGameMap.currLevel > 0)
                {
                  tetrisGameMap.currLevel--;
                }
                totalLineDone++;
                tetrisGameMap.currScore += Memory.SCORE_PER_LINE * totalLineDone;
                score.Invalidate();
                piecesPanel.Invalidate();
                break;
              }
              if (i == 0)
              {
                checkLinesEnded = true;
              }
            }
          }
        }
        else
        {
          isRelasedPiece = false;
          isArrived = false;
          checkingLines = false;
          checkLinesEnded = true;
          tetrisGameMap.currX = (Memory.SCREEN_COLUMNS - Memory.PEACE_SIZE_X) / 2;
          tetrisGameMap.currY = 0;
          tetrisGameMap.currPieceType = tetrisGameMap.nextType;
          tetrisGameMap.currPiece = (bool[,])tetrisGameMap.next.Clone();
          tetrisGameMap.currPieceRotation = Memory.pieceRotation.NO;
          tetrisGameMap.nextType = tetrisGameMap.getRandomPiece();
          tetrisGameMap.next = (bool[,])tetrisGameMap.getPiece(tetrisGameMap.nextType, Memory.pieceRotation.NO).Clone();
          nextPiece.Invalidate();
          piecesPanel.Invalidate();
          if (!controller.testCanMove(Controller.Directions.CheckStartPiece, tetrisGameMap))
          {
            isGameOver = true;

            /*salva la partita giocata su un file
             * using (Stream stream = File.Open(DateTime.Now.ToString("yyyy-MM-dd-hh-mm-ss") + ".bin", FileMode.Create))
            {
              var bformatter = new System.Runtime.Serialization.Formatters.Binary.BinaryFormatter();

              bformatter.Serialize(stream, recordGame);
            }
            */
            //scrive record su www.tuttosi.info
            /*
            string base64GameRecord = null;
            using (MemoryStream stream = new MemoryStream())
            {
              var bformatter = new System.Runtime.Serialization.Formatters.Binary.BinaryFormatter();
              bformatter.Serialize(stream, recordGame);
              string serialzedList = StreamToString(stream);
              string zippedList = Encoding.UTF8.GetString(Zip(serialzedList));
              base64GameRecord = Base64Encode(zippedList);
            }

            File.WriteAllText(DateTime.Now.ToString("yyyy-MM-dd-hh-mm-ss") + ".txt", base64GameRecord);
            WebClient myWebClient = new WebClient();
            NameValueCollection serviceParms = new NameValueCollection();
            serviceParms.Add("nome", "Antonio ama Valentina");
            serviceParms.Add("score", tetrisGameMap.currScore.ToString());
            serviceParms.Add("level", tetrisGameMap.currLevel.ToString());
            serviceParms.Add("lines", tetrisGameMap.currLines.ToString());
            serviceParms.Add("gameRecorded", base64GameRecord);
            myWebClient.UploadValues("http://www.tuttosi.info/cgi-bin/forms/scriviRecord.asp", serviceParms);
            */

            return;
          }
        }
      }
      timer1.Enabled = true;
    }

    private void cmdLoadDemo_Click(object sender, EventArgs e)
    {
      var assembly = Assembly.GetExecutingAssembly();
      var resourceName = "tetris.v4.Resources.2019-02-14-03-06-46.bin";
      using (Stream stream = assembly.GetManifestResourceStream(resourceName))
      {
        var bformatter = new System.Runtime.Serialization.Formatters.Binary.BinaryFormatter();
        recordGame = (List<Recorder>)bformatter.Deserialize(stream);
      }

      currTick = 0;
      isGameOver = false;
      isDemoMode = true;
      timer1.Enabled = true;

    }

    //public static string StreamToString(Stream stream)
    //{
    //  stream.Position = 0;
    //  using (StreamReader reader = new StreamReader(stream, Encoding.UTF8))
    //  {
    //    return reader.ReadToEnd();
    //  }
    //}
    
    //public static Stream StringToStream(string src)
    //{
    //  byte[] byteArray = Encoding.UTF8.GetBytes(src);
    //  return new MemoryStream(byteArray);
    //}

    //public static string Base64Encode(string plainText)
    //{
    //  var plainTextBytes = System.Text.Encoding.UTF8.GetBytes(plainText);
    //  return System.Convert.ToBase64String(plainTextBytes);
    //}

    //public static string Base64Decode(string base64EncodedData)
    //{
    //  var base64EncodedBytes = System.Convert.FromBase64String(base64EncodedData);
    //  return System.Text.Encoding.UTF8.GetString(base64EncodedBytes);
    //}

    //public static byte[] Zip(string str)
    //{
    //  var bytes = Encoding.UTF8.GetBytes(str);

    //  using (var msi = new MemoryStream(bytes))
    //  using (var mso = new MemoryStream())
    //  {
    //    using (var gs = new GZipStream(mso, CompressionMode.Compress))
    //    {
    //      //msi.CopyTo(gs);
    //      CopyTo(msi, gs);
    //    }

    //    return mso.ToArray();
    //  }
    //}

    //public static string Unzip(byte[] bytes)
    //{
    //  using (var msi = new MemoryStream(bytes))
    //  using (var mso = new MemoryStream())
    //  {
    //    using (var gs = new GZipStream(msi, CompressionMode.Decompress))
    //    {
    //      //gs.CopyTo(mso);
    //      CopyTo(gs, mso);
    //    }

    //    return Encoding.UTF8.GetString(mso.ToArray());
    //  }
    //}

    //public static void CopyTo(Stream src, Stream dest)
    //{
    //  byte[] bytes = new byte[4096];

    //  int cnt;

    //  while ((cnt = src.Read(bytes, 0, bytes.Length)) != 0)
    //  {
    //    dest.Write(bytes, 0, cnt);
    //  }
    //}
  }
}
